package net.hoopajoo.cheeselab3;
import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.sql.*;
import java.util.*;
import java.util.regex.*;
import similaritymetrics.*;

public class PlayerThread extends Thread {
	SocketChannel mud = null;
	SocketChannel sock = null;
	boolean running = true;
	int debuglevel = 0;
	
	// Future movement places
	Vector future = new Vector();
	
	boolean moved = false;
	long last_syncid = 0;
	boolean online = false;
	
	Pattern score_pattern = Pattern.compile( "\\d+/\\d+ hits," );
	Pattern info_pattern = Pattern.compile( "You have \\d+/\\d+ hit," );
	
	String room_color = null;
	String last_auto_room_color = null;
	
	// direction constants
	final int DIR_UNKNOWN = -1;
	final int DIR_NONE = 0;
	final int DIR_NORTH = 1;
	final int DIR_SOUTH = 2;
	final int DIR_EAST = 3;
	final int DIR_WEST = 4;
	final int DIR_UP = 5;
	final int DIR_DOWN = 6;
	
	// output process constants
	final int PROC_PROMPT = 0;
	final int PROC_ROOM_TITLE = 1;
	final int PROC_EDESC = 2;
	final int PROC_EXITS = 3;
	
	int last_output = PROC_PROMPT;
	
	boolean autosave = false,
		automerge = false,
		autoforce = false;
	
	int x = 10000,
		y = 10000,
		z = 10000;
	
	
	
	api.AbstractStringMetric metrics = new Levenstein();
	float min_similarity_pct = (float)0.98; // % required to match edesc/etc for moves
	float autosave_min_similarity_pct = (float)1.0; // % required to match edesc/etc in autosave
	
	String username;
	CheesePlayerData player_data;
	
	String following = null;
	Pattern following_ex = null;
	
	int hits = 0, hits_max = 0,
		moves = 0, moves_max = 0,
		mana = 0, mana_max = 0;
	
	// Move fails
	Vector movefails;
	    
    // intermediate line saves
    ByteBuffer last_partial_input = ByteBuffer.allocateDirect( 10 * 1024 );
    ByteBuffer last_partial_output = ByteBuffer.allocateDirect( 10 * 1024 );
    
    // In autosave mode we sychronize sending lines to the mud with prompts
    // because if you spam too fast, it will not output all rooms that you see
    // e.g. if MUME sees like nnnnn in your input queue, you might only get
    // 3 of those rooms returned as room titles.  Not a big deal in follow mode
    // but it sucks in map mode because you "lose" rooms that never get mapped
	ByteBuffer sync_io = ByteBuffer.allocateDirect( 10 * 1024 );
	
    // last_room is the last room *from the database*
	CheeseRoomData last_room = null;
	int last_move = DIR_UNKNOWN;
	// This is the current room *from the database*
	CheeseRoomData curr_room = null;
	
	// This is where we build up the data to find where we are
    CheeseRoomData current = new CheeseRoomData();
    boolean got_exits = false;

    Vector last_potential_rooms = null;
    
    Connection db;

	Vector movestack = new Vector();
	
	PlayerThread( SocketChannel s ) {
		sock = s;
		load_movefails();
	}
	
	private void load_movefails() {
		BufferedReader r;
		movefails = new Vector();
		try {
			r = new BufferedReader( new FileReader( "move_pop.txt" ) );
			String line;
			while( (line = r.readLine()) != null ) {
				if( line.startsWith( "#" ) ) {
					//comment
				}else{
					CheeseLab.dmesg( 2, "Adding failure: '" + line + "'" );
					Pattern p = Pattern.compile( line );
					movefails.add( p );
				}
			}
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}
	
	private void closeup() {
		try {
			CheeseLab.dmesg( 1,"Closing player socket" );
			sock.close();
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Unable to close player socket" );
		}
	}
	
	public void run() {
		CheeseLab.dmesg( 1, "Starting proxy thread" );
		
		last_partial_input.clear();
		last_partial_output.clear();
		sync_io.clear();
		
		// Get db
		try {
			/*
			Class.forName( "org.postgresql.Driver" );
			db = DriverManager.getConnection( "jdbc:postgresql://localhost/cheeselab3", "postgres", "postgres" );
			stmt = db.createStatement();
			*/
			db = DriverManager.getConnection( "jdbc:derby:cheesedb" );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Unable to load database: " + e );
			closeup();
			return;
		}
		
		// Greet
		try {
			println_player(	"CheeseLab MUME Mapping Tools v." + CheeseLab.version );
			println_player( "" );

			FileInputStream fstream = new FileInputStream( "greeting.txt" );
			BufferedReader greet = new BufferedReader( new InputStreamReader( fstream ) );
			String line;
			while( (line = greet.readLine() ) != null ) {
				println_player( line );
			}
			fstream.close();
			print_player( "\r\n" );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0,"Error printing greeting: " + e );
		}
		
		// If there is only 1 user, and his password is blank or null,
		// don't bother with auth
		boolean perform_auth = true;
		try {
			PreparedStatement s = db.prepareStatement( "select count(*) as c from user_t" );
			ResultSet rs = s.executeQuery();
			if( rs.next() ) {
				// Check pw
				int count = rs.getInt( "c" );
				s = db.prepareStatement( "select userid, password from user_t" );
				rs = s.executeQuery();
				if( rs.next() ) {
					String pw = rs.getString( "password" );
					if( pw == null )
						pw = "";
					
					if( pw.equals( "" ) && count == 1 ) {
						perform_auth = false;
						username = rs.getString( "userid" );
					}
				}else{
					CheeseLab.dmesg( 2, "Unable to check single user password?" );
					System.exit( 1 );
				}
			}else{
				CheeseLab.dmesg( 2, "No users?" );
				System.exit( 1 );
			}
			rs.close();
			s.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		if( perform_auth ) {
			// Auth
			try {
				println_player( "" );
				print_player( "Username: " );
				
				username = readline_player();
				
				print_player( "Password: " );
				ByteBuffer codes = ByteBuffer.allocate( 5 );
				byte[] echo_off = { (byte)255, (byte)251, 1 };
				byte[] echo_on = { (byte)255, (byte)252, 1 };
								
				codes.clear();
				sock.write( (ByteBuffer)codes.put( echo_off ).flip() );

				String password = "";
				while( password.equals( "" ) )
					password = strip_line( readline_player_bytes() );
				
				codes.clear();
				sock.write( (ByteBuffer)codes.put( echo_on ).flip() );

				// Check it out
				//password = password.substring( 3 ); // strip telnet response to echo off
				CheeseLab.dmesg( 3, "Validating: " + username + " with '" + password + "'" );
				PreparedStatement s = db.prepareStatement( "select password from user_t where userid=?" );
				s.setString( 1, username );
				ResultSet rs = s.executeQuery();
				boolean login_ok = false;
				if( rs.next() ) {
					// Username was valid, check pw
					if( password.equals( rs.getString( 1 ) ) ) {
						CheeseLab.dmesg( 3, "OK!" );
						login_ok = true;
					}
				}
				rs.close();
				s.close();
				
				if( ! login_ok ) {
					println_player( "Login failed" );
					closeup();
					return;
				}
			}catch( Exception e ) {
				CheeseLab.dmesg( 0, "Error during auth: " + e );
				e.printStackTrace();
				closeup();
				return;
			}
		}
		
		// After push update
		CheeseProxy.changed();
		
		// Connect to mume
		try {
			println_player( "Connecting to: " + CheeseProxy.props.getProperty( "server" ) + ":" + CheeseProxy.props.getProperty( "port" ) );
			
			mud = SocketChannel.open();
			mud.connect( new InetSocketAddress( CheeseProxy.props.getProperty( "server" ), Integer.parseInt( CheeseProxy.props.getProperty( "port" ) ) ) );
			//mud.connect( new InetSocketAddress( "mume.pvv.org", 23 ) );
			//mud.connect( new InetSocketAddress( "localhost", 4242 ) );			
		}catch( Exception e ) {
			println_player( "Unable to connect to mud: " + e );
			CheeseLab.dmesg( 0, "Unable to connect to mud: " + e );
			e.printStackTrace();
			closeup();
			return;
		}
		
		// Yes to loop
		online = true;
		
		// Setup our selector and move all to non-blocking
		Selector s = null;
		SelectionKey mud_key = null,
			sock_key = null;
		try {
			s = Selector.open();
			sock.configureBlocking( false );			
			mud.configureBlocking( false );
			mud_key = mud.register( s, SelectionKey.OP_READ );
			sock_key = sock.register( s, SelectionKey.OP_READ );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Error setting up selector: " + e );
			e.printStackTrace();
		}
		
		// Main loop
		CheeseLab.dmesg( 2, "Starting main player loop" );
		ByteBuffer buf = ByteBuffer.allocateDirect( 100 * 1024 );
		while( online ) {
			try {
				int num = s.select();
				
				if (num == 0) {
				      continue;
			   }

			   Set keys = s.selectedKeys();
			   Iterator it = keys.iterator();

			   while (it.hasNext()) {
			      SelectionKey key = (SelectionKey)it.next();
			      SocketChannel sc = (SocketChannel)key.channel();
			      
			      buf.clear();
			      if( sc.read( buf ) < 1 ) {
			      	CheeseLab.dmesg( 2, "Read error, connection closed" );
			      	online = false;
			      }else{
			      	CheeseLab.dmesg( 3, "Got " + buf.remaining() + " bytes left" );
			      	CheeseLab.dmesg( 3, "  Limit is: " + buf.limit() );			      	
			      
				      buf.flip();
				      			      
				      if( key == sock_key ) {
				      	CheeseLab.dmesg( 3, "Printing to mud" );
				      	mud.write( process_input( buf ) );
				      	//mud.socket().getOutputStream().flush();
				      }else{
				      	CheeseLab.dmesg( 3, "Printing to player" );
				      	sock.write( process_output( buf ) );
				      	//sock.socket().getOutputStream().flush();
				      }
			      }
			   }

			   keys.clear();
			}catch( Exception e ) {
				CheeseLab.dmesg( 0, "Error in main loop: " + e );
				e.printStackTrace();
			}
		}
		
		online = false;
		save_player();

		try {
			s.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		try {
			mud.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		// Done
		CheeseLab.dmesg( 2, "Closing server: " + this );
		closeup();
	}
	
	private ByteBuffer process_output( ByteBuffer b ) {
		last_partial_output.put( b );
		byte[] buff = null;
		boolean looping = true;
		while( looping ) {
			//CheeseLab.dmesg( "processing buffer: " + new String( last_partial_output.array() ) );
			CheeseLab.dmesg( 3, "Buffer has " + last_partial_output.remaining() + " bytes left" );
			last_partial_output.flip();
			buff = next_raw_line( last_partial_output );
			last_partial_output.compact();
			
			if( buff.length == 0 || buff[ buff.length - 1 ] != '\n' ) {
				if( buff.length != 0 && buff[ buff.length - 1 ] == '>' ) {
					// this is a prompt, it passes this back as a "line" even
					// though it does not end with \n
					CheeseLab.dmesg( 4, "Got prompt (i think), discarding: " + new String( buff ) );
				
					// We still process the prompt but we do not carry it over
					process_output_line( buff );			
					buff = new byte[ 0 ];
				}else{
					CheeseLab.dmesg( 4, "Got partial, saving for later: " + new String( buff ) );
				}
				//CheeseLab.dmesg( "Discarding partial line: " + new String( buff ) );
				looping = false;
			}else{
				//buff[ buff.length - 1 ] = 0;
				CheeseLab.dmesg( 4, " >> Got line: " + new String( buff ) );
				
				// Now for incoming logic!!
				process_output_line( buff );
			}
		}
		last_partial_output.clear();
		last_partial_output.put( buff );
		
		/* 
		 for( int i = 0; i < b.remaining(); i++ ) {
			CheeseLab.dmesg( "Char at " + i + " = " + buff[ i ] );
			if( buff[ i ] == 27 ) {
				CheeseLab.dmesg( "Escape!!" );
			}
		}
		
		try {
			//CheeseLab.dmesg( new String( b.array(), "UTF-8" ) );
			//System.out.write( b.array(), 0, b.remaining() );
			
			String gonnathing = new String( b.array(), "ISO-8859-1" );
			PrintWriter out = 
			      new PrintWriter(
			          new OutputStreamWriter(System.out,
			                                 "ISO-8859-1"));
			  out.println( gonnathing );
		}catch( Exception e ){
	
		}
		*/
		
		b.flip();
		return( b );
	}
	
	private void process_output_line( byte[] l ) {
		if( l.length == 0 )
			return;
		
		byte[] escape = { 27 };
		String str_escape = new String( escape );
		String room_prefix = str_escape + "[";
		if( room_color != null ) 
			room_prefix += room_color;
		
		String str = new String( l );
		String clean_str = new String( strip_line( l ) );
		
		// we should verify this is not an obvious tell or something
		if( ( str.startsWith( room_prefix ) ||
				str.startsWith( "It is pitch black..." ) ||
				str.startsWith( "You just see a dense fog around you" ) )
				// make sure it's not a tell or narrate (which may be proc'd in the
				// case of auto-room detection
				&& str.indexOf( "tells you '" ) == -1
				&& str.indexOf( "narrates '" ) == -1 ) {
			// the room name (if colors are set right) or a new room
			// that you cannot see in
			
			// Save the color code if we have not detected the color code
			if( room_color == null && str.startsWith( room_prefix ) ) {
				if( str.length() > 4 ) {
					last_auto_room_color = str.substring( 2, str.indexOf( "m" ) + 1 );
					cdebug( 3, "Saving auto color: " + last_auto_room_color );
				}
			}
			
			// Reset our info collector
			reset_current();
			
			cdebug( 4, "This line is a room name" );
			if( str.startsWith( room_prefix ) )
				current.name = strip_line( l );
			
			last_output = PROC_ROOM_TITLE;
			moved = true;
		}else if( l[ l.length - 1 ] == '>' ) {
			cdebug( 4, "This line is a prompt" );
			last_output = PROC_PROMPT;
			
			// Terrain is 2nd char
			if( clean_str.length() > 2 ) { // dunno why this would not be true
				current.terrain = translate_terrain( clean_str.substring( 1, 2 ) );
			}
			
			// Locate
			find_me();
			
			// Send sync io events if the are queued
			if( sync_io.position() > 0 ) {
				CheeseLab.dmesg( 4, "Passing synchro io" );
				sync_io.flip();
				try {
					mud.write( ByteBuffer.wrap( next_raw_line( sync_io  ) ) );
				}catch( Exception e ) {
					e.printStackTrace();
				}
				sync_io.compact();
			}
		}else if( last_output == PROC_ROOM_TITLE  && l.length > 2 && l[ l.length - 2 ] != '\r' ) {
			cdebug( 4, "This line is the first edesc line" );
			current.edesc = strip_line( l );
			last_output = PROC_EDESC;
		}else if( last_output == PROC_EDESC && l.length > 2 && l[ l.length - 2 ] != '\r' ) {
			// edesc continuation
			cdebug( 4, "This line is a room edesc continuation" );
			current.edesc += "\n" + strip_line( l );
			last_output = PROC_EDESC;
		}else if( clean_str.startsWith( "Exits: " ) ) {
			cdebug( 4, "This is the exits line" );
			got_exits = true;
			
			// parse exits and flag our room
			String just_exits = clean_str.substring( 6 );
			String[] exits = just_exits.split( "," );
			for( int i = 0; i < exits.length; i++ ) {
				int starts = 0;
				boolean door = false;
				int dir = DIR_UNKNOWN;
				// dirs are nsewud where n==1, s==2, etc
				
				CheeseLab.dmesg( 4, "Processing exit: " + exits[ i ] );
				while( dir == DIR_UNKNOWN ) {
					char c = exits[ i ].charAt( starts++ );
					
					if( c == '[' )
						door = true;
					
					if( c == '(' ) 
						door = true;
					
					// TODO: fix it
					//dir = direction_to_enum( new String( c ) );
					if( c == 'n' )
						dir = DIR_NORTH;
					
					if( c == 's' )
						dir = DIR_SOUTH;
					
					if( c == 'e' )
						dir = DIR_EAST;
					
					if( c == 'w' )
						dir = DIR_WEST;
					
					if( c == 'u' )
						dir = DIR_UP;
					
					if( c == 'd' )
						dir = DIR_DOWN;
					
					if( starts > exits[ i ].length() )
						dir = DIR_NONE;
				}
				
				switch( dir ) {
					case DIR_NORTH:
						current.exit_n = -1;
						current.door_n = door ? "exit" : "";
						break;
						
					case DIR_SOUTH:
						current.exit_s = -1;
						current.door_s = door ? "exit" : "";
						break;

					case DIR_EAST:
						current.exit_e = -1;
						current.door_e = door ? "exit" : "";
						break;

					case DIR_WEST:
						current.exit_w = -1;
						current.door_w = door ? "exit" : "";
						break;

					case DIR_UP:
						current.exit_u = -1;
						current.door_u = door ? "exit" : "";
						break;

					case DIR_DOWN:
						current.exit_d = -1;
						current.door_d = door ? "exit" : "";
						break;
				}
				CheeseLab.dmesg( 4, " > Got: " + dir + " (" + door + ")" );
			}
			
			last_output = PROC_EXITS;
		}else{
			// Check for failed moves
			for( int i = 0; i < movefails.size(); i++ ) {
				Pattern p = (Pattern)movefails.get( i );
				if( p.matcher( clean_str ).lookingAt() ) {
					cdebug( 3, "Detected movement failure, pop!" );
					if( movestack.size() > 0 )
						movestack.remove( 0 );
					
					cdebug( 5, "Movestack:" );
					for( int ii = 0; ii < movestack.size(); ii++ ) {
						//cdebug( 5, "  -> " + movestack.get( ii ) );
						cdebug( 5, "  -> " + movestack.get( ii ) + " - " + enum_to_dir( ((Integer)movestack.get( ii )).intValue() ) );
					}

					i = movefails.size(); // break
				}
			}
			
			// check for flee
			if( clean_str.startsWith( "You flee head over heels." ) ) {
				CheeseLab.dmesg( 3, "Flee!" );
				movestack.clear();
				movestack.add( new Integer( DIR_UNKNOWN ) );
			}
			
	        // following
			if( clean_str.startsWith( "You now follow" ) ) {
				// Extract name (passed follow and before punctuation)
				following = clean_str.substring( "You now follow ".length(),
						clean_str.length() - 1 );
				
				// make a non-labeled version for matches because
				// when someone is riding you do not see their label
				// when they move
				String clean_fol = following;
				if( following.indexOf( "(" ) > 0 ) {
					clean_fol = following.substring( 0, following.indexOf( "(" ) - 1 );
				}
				cdebug( 3, "Following: " + following );
				cdebug( 4, "Clean name is '" + clean_fol + "'" );
				
				// setup a regex we can use for this later
				clean_fol = clean_fol.toLowerCase();
				following_ex = Pattern.compile( clean_fol + " leaves" );
			}
			
			if( clean_str.startsWith( "You will not follow anyone else now." ) ) {
				cdebug( 3, "Stopped following" );
				following = null;
				following_ex = null;
			}
				        
	        // leader moving
			if( following != null ) {
				String lower = clean_str.toLowerCase(); 
				Matcher m = following_ex.matcher( lower );

				if( m.lookingAt() ) {
					cdebug( 3, "Leader moved" );
					
					String getdir = clean_str.substring( clean_str.indexOf( "leaves " ) +
							"leaves ".length() );
					if( getdir.indexOf( " " ) > 0 ) {
						// Riding
						getdir = getdir.substring( 0, getdir.indexOf( " " ) );
					}else{
						// Trim the puncuation
						getdir = getdir.substring( 0, getdir.length() - 1 );
					}
					
					cdebug( 4, "Following: " + getdir );
					movestack.clear(); // if following don't use multiple moves
					movestack.add( new Integer( direction_to_enum( getdir ) ) );
				}
			}
			
			// Typed SCO (note that this will note work if LOOK color is same as STAT color
			if( score_pattern.matcher( clean_str ).lookingAt() ||
					info_pattern.matcher( clean_str ).lookingAt() ) {
				CheeseLab.dmesg( 3, "Ripping player score" );
				String[] parts = clean_str.split( "," );
				
				hits = hits_max = mana = mana_max = moves = moves_max = 0;
				
				for( int i = 0; i < parts.length; i++ ) {
					// extract digits
					int[] n = extract_digit_parts( parts[ i ] );
					if( parts[ i ].indexOf( "hit" ) != -1 ) {
						hits = n[ 0 ];
						hits_max = n[ 1 ];
					}else if( parts[ i ].indexOf( "mana" ) != -1 ) {
						mana = n[ 0 ];
						mana_max = n[ 1 ];						
					}else if( parts[ i ].indexOf( "move" ) != -1 ) {
						moves = n[ 0 ];
						moves_max = n[ 1 ];
					}
				}
				save_player();
			}			
		}
		
		//print_current();
	}
	
	private int[] extract_digit_parts( String s ) {
		int[] p = new int[ 2 ];
		// In theory if we removee anything that is NOT a number, 
		// we have part 1 and 2 now
		String[] parts = s.replaceAll( "[^0-9/]", "" ).split( "/" );

		p[ 0 ] = Integer.parseInt( parts[ 0 ] );
		p[ 1 ] = Integer.parseInt( parts[ 1 ] );
		
		return( p );
	}
	
	private ByteBuffer process_input( ByteBuffer b ) {
		last_partial_input.put( b );
		b.clear();
		byte[] buff = null;
		boolean looping = true;
		
		while( looping ) {
			//CheeseLab.dmesg( "processing buffer: " + new String( last_partial_output.array() ) );
			last_partial_input.flip();
			buff = next_raw_line( last_partial_input );
			last_partial_input.compact();
			
			if( buff.length == 0 || buff[ buff.length - 1 ] != '\n' ) {
				//CheeseLab.dmesg( "Discarding partial line: " + new String( buff ) );
				looping = false;
			}else{
				//buff[ buff.length - 1 ] = 0;
				CheeseLab.dmesg( 4, "Got input line: " + new String( buff ) );
				
				// Now for incoming logic!!
				b.put( process_input_line( buff ) );
			}
		}
		last_partial_input.clear();
		last_partial_input.put( buff );

		// If autosave, then queue this for sending via the prompt
		if( autosave ) {
			b.flip();
			CheeseLab.dmesg( 4, "Queuing syncro io: " + b.limit() );
			
			boolean pass_1_line = false;
			if( sync_io.position() == 0 )
				// there is nothing in the syncro cache so we need to pass at least 1 line
				pass_1_line = true;
				
			sync_io.put( b );
			b.clear();
			
			// If pass_1_line the pull a line and pass it
			if( pass_1_line ) {
				CheeseLab.dmesg( 3, "Sync buffer was empty, passing 1 line" );
				sync_io.flip();
				b.put( next_raw_line( sync_io ) );
				sync_io.compact();
			}
		}
		
		b.flip();
		return( b );
	}
	
	private byte[] process_input_line( byte[] l ) {
		byte[] ret;
		
		String input = new String( strip_line( l ) );
		CheeseLab.dmesg( 4, "Processing input line: " + input );
		if( l[ 0 ] == '@' ) {
			String lower = input.toLowerCase();
			
			// Get our user info for permissions checks
			player_data = CheeseDecoder.get_player( username, db );
			
			CheeseLab.dmesg( 4, "Server command: " + input );
			if( lower.startsWith( "@save" ) && player_data.can_save() ) {
				// Save room
				cmd_save( input );
			}else if( lower.startsWith( "@serverlog" ) && player_data.is_administrator() ) {
				// Change server log level
				String[] parts = lower.split( " " );
				if( parts.length > 1 ) {
					try {
						CheeseLab.log_level = Integer.parseInt( parts[ 1 ] );
						println_player( "Log level changed" );
					}catch( Exception e ) {
						println_player( "Unable to parse log level" );
					}
				}else{
					println_player( "Invalid log level" );
				}
			}else if( lower.startsWith( "@autosave" ) && player_data.can_save() ) {
				String[] split = lower.split( " " );
				autosave = automerge = autoforce = false;
				if( split.length < 2 ) {
					// no args
					autosave = true;
				}else{
					// handle shutting it off
					if( split[ 1 ].equals( "off" ) ) {
						autosave = automerge = autoforce = false;
					}else if( split[ 1 ].equals( "merge" ) ) {
						// or auto-merge mode too
						autosave = true;
						automerge = true;
					}else if( split[ 1 ].equals( "force" ) ) {
						autosave = true;
						autoforce = true;
					}else{
						autosave = true;
					}
				}
			}else if( lower.startsWith( "@door" ) && player_data.can_save() ) {
				if( curr_room != null ) {
					String[] split = lower.split( " " );
					int dir = direction_to_enum( split[ 1 ] );
					String dstring = null;
					String estring = null;
					int to_id = 0;
					
					switch( dir ) {
						case DIR_NORTH:
							dstring = "door_n";
							estring = "exit_n";
							to_id = curr_room.exit_n;
							break;
						case DIR_SOUTH:
							dstring = "door_s";
							estring = "exit_s";
							to_id = curr_room.exit_s;
							break;
						case DIR_EAST:
							dstring = "door_e";
							estring = "exit_e";
							to_id = curr_room.exit_e;
							break;
						case DIR_WEST:
							dstring = "door_w";
							estring = "exit_w";
							to_id = curr_room.exit_w;
							break;
						case DIR_UP:
							dstring = "door_u";
							estring = "exit_u";
							to_id = curr_room.exit_u;
							break;
						case DIR_DOWN:
							dstring = "door_d";
							estring = "exit_d";
							to_id = curr_room.exit_d;
							break;
					}
					
					if( dstring != null  ) {
						try {
							String exitname;
							if( split.length < 3 )
								exitname = ""; // clear
							else
								exitname = split[ 2 ];

							PreparedStatement update;
							
							// If the exit is 0 then make it -1
							if( to_id == 0 ) {
								update = db.prepareStatement( "update room_t set " + estring + "=? where roomid=?" );
								update.setInt( 1, -1 );
								update.setInt( 2, curr_room.roomid );
								update.executeUpdate();
							}

							// Update exit and mtime/etc
							update = db.prepareStatement( "update room_t set " + dstring + "=?, mtime=?, userid=? where roomid=?" );
							update.setString( 1, exitname );
							update.setLong( 2, System.currentTimeMillis() );
							update.setString( 3, username );
							update.setInt( 4, curr_room.roomid );
							update.executeUpdate();
							update.close();
							
							CheeseProxy.world_changed();
						}catch( Exception e ){
							e.printStackTrace();
						}
					}
				}
			}else if( lower.startsWith( "@notes" ) && player_data.can_save() ) {
				// remove prefix
				if( curr_room != null ) {
					try {
						String notes = lower.substring( "@notes ".length() );
						PreparedStatement update = db.prepareStatement( "update room_t set notes=?, mtime=?, userid=? where roomid=?" );
						update.setString( 1, notes );
						update.setLong( 2, System.currentTimeMillis() );
						update.setString( 3, username );
						update.setInt( 4, curr_room.roomid );
						update.executeUpdate();
						update.close();
						
						CheeseProxy.world_changed();
					}catch( Exception e ) {
						e.printStackTrace();
					}
				}				
			}else if( lower.equals( "@delete" ) && player_data.can_save() ) {
				// delete curr
				if( curr_room != null ) {
					try {
						PreparedStatement update = db.prepareStatement( "delete from room_t where roomid=?" );
						update.setInt( 1, curr_room.roomid );
						update.executeUpdate();
						update.close();
						
						String[] dirs = { "n", "s", "e", "w", "u", "d" };
						for( int i = 0; i < dirs.length; i++ ) {
							update = db.prepareStatement( "update room_t set exit_" + dirs[ i ] + "=-1, userid=?, mtime=? where exit_" + dirs[ i ] + "=?" );
							update.setString( 1, username );
							update.setLong( 2, System.currentTimeMillis() );
							update.setInt( 3, curr_room.roomid );

							update.executeUpdate();
							update.close();
						}
						
						curr_room = null;
						
						CheeseProxy.world_changed();
					}catch( Exception e ) {
						e.printStackTrace();
					}
				}
			}else if( lower.equals( "@stack" ) ) {
				cmd_stack();
			}else if( lower.startsWith( "@debug" ) ) {
				String[] parts = lower.split( " " );
				if( parts.length > 1 ) {
					try {
						debuglevel = Integer.parseInt( parts[ 1 ] );
					}catch( Exception e ) {
						println_player( "Unable to parse debug level" );
					}
				}
			}else if( lower.equals( "@help" ) ) {
				cmd_help();
			}else if( lower.startsWith( "@user" ) && player_data.can_modify_users() ) {
				cmd_user( input );
			}else if( lower.startsWith( "@password" ) ) {
				cmd_password( input );
			}else if( lower.equals( "@reset" ) ) {
				movestack.clear();
			}else if( lower.startsWith( "@overlay" ) && player_data.can_save() ) {
				String overlay = lower.substring( "@overlay".length() ).trim();
				if( curr_room != null && ! overlay.equals( "" ) ) {
					// you can overlay tile or remove with -tile
					
					// We make a hashmap of the overlays, then join the keys
					// together so we don't have dupes
					HashMap unique = new HashMap();
					for( int i = 0; i < curr_room.overlays.length; i++ ) {
						unique.put( curr_room.overlays[ i ], new Integer( 1 ) );
					}
					
					if( overlay.startsWith( "-" ) ) {
						// remove
						overlay = overlay.substring( 1 );
						CheeseLab.dmesg( 3, "Removing overlay: " + overlay );
						unique.remove( overlay );
					}else{
						CheeseLab.dmesg( 3, "Adding overlay: " + overlay );
						unique.put( overlay, new Integer( 1 ) );
					}
					
					String overlays = "";
					Object[] oi = unique.keySet().toArray();
					
					for( int i = 0; i < oi.length; i++ ) {
						if( i > 0 )
							overlays += ",";
						overlays += (String)oi[ i ];
					}

					CheeseLab.dmesg( 3, "Saving overlays: " + overlays );
					// And finally save
					try {
						PreparedStatement update = db.prepareStatement( "update room_t set overlays=?, mtime=?, userid=? where roomid=?" );
						update.setString( 1, overlays );
						update.setLong( 2, System.currentTimeMillis() );
						update.setString( 3, username );
						update.setInt( 4, curr_room.roomid );
						update.executeUpdate();
						update.close();
						
						CheeseProxy.world_changed();
					}catch( Exception e ) {
						e.printStackTrace();
					}
				}
			}else{
				// Unknown command
				println_player( "Unknown command or invalid permissions for: " + lower );
			}
			
			ret = new byte[ 0 ]; // do not pass
		}else{
			ret = l; // pass thru
			
			if( input.length() > 0 ) {
				// Check for movement
				int dir = direction_to_enum( input.trim() );
				if( dir != DIR_UNKNOWN ) {
					CheeseLab.dmesg( 3, "Saving move: " + dir );
					movestack.add( new Integer( dir ) );
					
					// Based on last vector move thing, try to see if we'd be able to go that way
					// and save it in futures for map rending of future paths
					CheeseRoomData r = curr_room;
					if( future.size() > 0 ) {
						String t = (String)future.get( future.size() - 1 );
						String[] parts = t.split( "," );
						int roomid = Integer.parseInt( parts[ 0 ] );
						if( roomid != 0 ) {
							r = CheeseDecoder.get_room( roomid, db );
						}
					}
					
					if( r == null ) {
						// we can't extract moves, maybe someday just overlay based on xyz
						CheeseLab.dmesg( 3, "Unable to make future vector, no room here" );
					}else{
						// Get the room we'd go to
						int toid = 0;
						switch( dir ) {
							case DIR_NORTH:
								toid = r.exit_n;
								break;

							case DIR_SOUTH:
								toid = r.exit_s;
								break;
								
							case DIR_EAST:
								toid = r.exit_e;
								break;
								
							case DIR_WEST:
								toid = r.exit_w;
								break;
								
							case DIR_UP:
								toid = r.exit_u;
								break;
								
							case DIR_DOWN:
								toid = r.exit_d;
								break;							
						}
						
						if( toid == -1 ) {
							CheeseLab.dmesg( 3, "Unable to do future: unexplored" );
						}else if( toid == 0 ) {
							CheeseLab.dmesg( 3, "Unable to do future: wall" );
						}else{
							// Grab fut room
							CheeseLab.dmesg( 3, "Added future room " + r.roomid );
							r = CheeseDecoder.get_room( toid, db );
							if( r != null ) {
								future.add( r.roomid + "," +
									r.x + "," +
									r.y + "," +
									r.z 
								);
							}
						}
					}
					
				}
				
				
				/* strict movecheck mode was supposed to allow riding behind somenoe, but it was very buggy and error prone
				 * so I decided to get rid of it and use the old method, so no ride behind but a better stack processor overall
				 * 
				else if( strict_movechecks ) { // if in strict mode we must catch look/scout/etc because room title == movepop == dir_unknown
					// check for look/exa
					String[] split = input.trim().split( " " );
					if( split.length == 1 ) {
						if( "look".startsWith( split[ 0 ] ) ||
								"examine".startsWith( split[ 0 ] ) ) {
							CheeseLab.dmesg( 3, "Caught look/exa, adding move_none" );
							movestack.add( new Integer( DIR_NONE ) );
						}
					}else{					
						if( "scout".startsWith( split[ 0 ] ) ) {
							// They are scouting
							CheeseLab.dmesg( 3, "Caught scouting" );
							movestack.add( new Integer( DIR_NONE ) );
						}
					}
				}
				*/
				
				CheeseLab.dmesg( 4, "Movestack:" );
				for( int i = 0; i < movestack.size(); i++ ) {
					CheeseLab.dmesg( 4, "  -> " + movestack.get( i ) );
				}
			}
		}

		return( ret );
	}
	
	private int direction_to_enum( String dir ) {
		String l = dir.toLowerCase();

		CheeseLab.dmesg( 4, "Extracting direction from: " + dir );
		
		if( l.equals( "" ) )
			return( DIR_UNKNOWN );
		
		if( "north".startsWith( l ) )
			return( DIR_NORTH );
		
		if( "south".startsWith( l ) )
			return( DIR_SOUTH );
		
		if( "east".startsWith( l ) )
			return( DIR_EAST );
		
		if( "west".startsWith( l ) )
			return( DIR_WEST );
		
		if( "up".startsWith( l ) )
			return( DIR_UP );
		
		if( "down".startsWith( l ) )
			return( DIR_DOWN );
		
		return( DIR_UNKNOWN );
	}
	
	private byte[] next_raw_line( ByteBuffer b ) {
		byte here = 0;
		byte[] ret;
		int i;
		
		CheeseLab.dmesg( 4, "for next_raw_line, buffer limit is: " + b.limit() );
		
		b.mark();
		for( i = 0; b.hasRemaining() && here != '\n' && here != '>'; i++ ) {
			here = b.get();
			//CheeseLab.dmesg( "Char at " + i + " is " + here );
		}

		CheeseLab.dmesg( 4, "Extracting line, length is: " + i );
		
		b.reset();
		ret = new byte[ i ];
		b.get( ret, 0, i );
				
		return( ret );
	}
	
	private void reset_current() {
		current = new CheeseRoomData();
		got_exits = false;
	}
		
	// basically return the buffer as a line without the ansi and \r or \n
	private String strip_line( byte[] s ) {
		byte[] b = new byte[ s.length ];
		int bi = 0;
		boolean in_ansi = false,
			in_telnet = false,
			telnet_go_to_se = false;
		
		
		// Also strip telnet protocol, IAC is start,
		// if followed by SB then go til SE, else
		// it's a 3 byte seq
		byte IAC = (byte)255;
		byte SB = (byte)250;
		byte SE = (byte)240;
		int telnet_count = 0;
		
		for( int i = 0; i < s.length; i++ ) {
			//Byte bbb = new Byte( s[ i ] );
			//CheeseLab.dmesg( "Stripping: " + bbb );
			
			// Crappy ansi detection
			if( s[ i ] == 27 && in_ansi == false ) {
				in_ansi = true;
			}
		
			// Detect telnet protocol
			if( s[ i ] == IAC && in_telnet == false ) {
				telnet_count = 0;
				in_telnet = true;
				telnet_go_to_se = false;
			}
			
			// Remove \r
			if( ! in_telnet && ! in_ansi && s[ i ] != '\r' && s[ i ] != '\n' ) {
				// detect backspace/delete ascii and remove chars
				if( s[ i ] == 127 || s[ i ] == 8 )
					bi--;
				else
					b[ bi++ ] = s[ i ];
			}

			// If in telnet protocol, detect endings
			if( in_telnet ) {
				telnet_count++;
				
				if( telnet_count == 2 ) {
					// second char, if it's SB then we go to SE
					if( s[ i ] == SB ) {
						telnet_go_to_se = true;
					}
				}else if( telnet_count == 3 && telnet_go_to_se == false ) {
					// End
					in_telnet = false;
				}else if( telnet_go_to_se ) {
					if( s[ i ] == SE )
						in_telnet = false;
				}
			}
			
			// Detect end of crappy ansi detection
			if( in_ansi == true && s[ i ] == 'm' ) {
				in_ansi = false;
			}			
		}
		
		String ret;
		try {
			// Set charset for funny symbols to be saved right
			ret = new String( b, 0, bi, "ISO-8859-1" );
		}catch( Exception e ) {
			System.out.println( "Unable to use encoding" + e );
			System.exit( 1 );
			
			// shut up the fucking compiler, but you will NEVER get here
			ret = "";
		}
		return( ret );
	}
	
	private void find_me() {
		CheeseLab.dmesg( 2, "Begin find_me()" );
		
		// We do this in case we are spamming somewhere and a prompt is placed in between new
		// room entries
		if( ! moved )
			return;
		moved = false;
		
		cdebug( 5, "Movestack before extraction:" );
		for( int i = 0; i < movestack.size(); i++ ) {
			cdebug( 5, "  -> " + movestack.get( i ) + " - " + enum_to_dir( ((Integer)movestack.get( i )).intValue() ) );
		}
		
		if(  movestack.isEmpty() )
			return;
			
		last_move = ((Integer)movestack.remove( 0 )).intValue();
		
		// Also pull of potential places to be
		if( ! future.isEmpty() )
			future.remove( 0 );
		
		if( last_move == DIR_UNKNOWN ) {
			// disable autosave
			autosave = autoforce = automerge = false;
			
			// Clear futures
			future.clear();
		}
		
		// if dir == dir_none then this was just the effect of a look or examine
		if( last_move == DIR_NONE ) {
			CheeseLab.dmesg( 3, "They looked, didn't move" );
			return;
		}
		
		// hmm we should be looking at curr_room since we set that
		last_room = curr_room;
		if( last_room == null ) {
			cdebug( 3, "Last room is NULL" );
		}else{
			cdebug( 3, "Last room is: " + last_room.roomid );
		}
		
		// Get move
		cdebug( 3, "Move is: " + last_move );
		boolean last_move_unexplored = false; // for mapping, this will allow
		                                      // the system to sync and allow save/merge
		                                      // but also save/force in the new "would be"
		                                      // location.  Without this, you have to
		                                      // save/force, then move the room out from
		                                      // under the dup place
		Vector potential_rooms = new Vector();
		
		// First we intialize the potential rooms list
		// this is then refined later, based on the information
		// available which may vary depending on following/lighting/flee/etc
		//
		// Make if we have a last room and the dir is explored, simply move us
		// there (unless we fled, which is DIR_UNKNOWN)
		PreparedStatement s = null; 
		
		if( last_room != null ) {
			if( last_move == DIR_UNKNOWN ) {
				// We fled, so make a new potential rooms list based on the exits
				// of our last room
				try {
					s = db.prepareStatement( 
							"select * from room_t where " +
							"	roomid=? OR roomid=? OR" + 
							"	roomid=? OR roomid=? OR" + 
							"	roomid=? OR roomid=? " ); 
					s.setInt( 1, last_room.exit_n );
					s.setInt( 2, last_room.exit_s );
					s.setInt( 3, last_room.exit_e );
					s.setInt( 4, last_room.exit_w );
					s.setInt( 5, last_room.exit_u );
					s.setInt( 6, last_room.exit_d );
				}catch( Exception e ) {
					e.printStackTrace();
				}
			}else{
				// If we knew our last room and went in a direction, then
				// go that way
				int roomid = 0;
				switch( last_move ) {
					case DIR_NORTH:
						roomid = last_room.exit_n;
						break;
						
					case DIR_SOUTH:
						roomid = last_room.exit_s;
						break;
						
					case DIR_EAST:
						roomid = last_room.exit_e;
						break;
						
					case DIR_WEST:
						roomid = last_room.exit_w;
						break;
						
					case DIR_UP:
						roomid = last_room.exit_u;
						break;
						
					case DIR_DOWN:
						roomid = last_room.exit_d;
						break;
				}
				
				if( roomid > 0 ) {
					try {
						s = db.prepareStatement( "select * from room_t where roomid=?" );
						s.setInt( 1, roomid );
					}catch( Exception e ) {
						e.printStackTrace();
					}
				}else{
					last_move_unexplored = true;
				}
			}
		}
		
		if( s == null ) {
			// Start up a potential rooms list
			// we did now know where we were, so make up a query
			// to accept the rooms based on what we know (room title)
			// and add in any potential last rooms info
			try {
				s = db.prepareStatement( "select * from room_t where name=?" ); 
				s.setString( 1, current.name );
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}
		
		// Pull db rooms
		try {
			ResultSet rs = s.executeQuery();
			while( rs.next() ) {
				potential_rooms.add( CheeseDecoder.sql_to_room( rs ) );
			}
			rs.close();
			s.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		// Now based on the potential rooms we could be in, eliminate the ones that
		// we are not in
		for( int i = 0; i < potential_rooms.size(); i++ ) {
			CheeseRoomData r = (CheeseRoomData)potential_rooms.get( i );
			boolean is_bad = false;
			
			cdebug( 3, "Checking: " + r.roomid );
			
			// If we had potential last rooms, validate against those
			if( last_potential_rooms != null ) {
				boolean found_match = false;
				
				// loop over all last potentials, at least one of them has to point here
				// or we assume here is not a possible place to be
				for( int t = 0; t < last_potential_rooms.size() && ! found_match; t++ ) {
					CheeseRoomData test = (CheeseRoomData)last_potential_rooms.get( t );
					switch( last_move ) {
						case DIR_NORTH:
							if( test.exit_n == 0 || test.exit_n == r.roomid )
								found_match = true;
							break;
							
						case DIR_SOUTH:
							if( test.exit_s == 0 || test.exit_s == r.roomid )
								found_match = true;
							break;
							
						case DIR_EAST:
							if( test.exit_e == 0 || test.exit_e == r.roomid )
								found_match = true;
							break;
							
						case DIR_WEST:
							if( test.exit_w == 0 || test.exit_w == r.roomid )
								found_match = true;
							break;
															
						case DIR_UP:
							if( test.exit_u == 0 || test.exit_u == r.roomid )
								found_match = true;
							break;
							
						case DIR_DOWN:
							if( test.exit_d == 0 || test.exit_d == r.roomid )
								found_match = true;
							break;
						
						case DIR_UNKNOWN:
							// flee, we allow all rooms after a flee
							found_match = true;
							break;
							
						default:
							break;
					}
				}
				
				// None of the potential rooms comes here
				if( ! found_match ) {
					cdebug( 3, " > Failed linked to last potential rooms check" );
					is_bad = true;
				}
			}
			
			float req_similarity_pct = min_similarity_pct;
			if( autosave ) {
				req_similarity_pct = autosave_min_similarity_pct;
			}
			// Do main things, like edesc/name/terrain/etc
			if( ! current.edesc.equals( "" ) ) {
				// Do similar
				float simdex = metrics.getSimilarity( current.edesc, r.edesc );
				cdebug( 3, "Edesc similarity is: " + simdex );
				
				if( simdex < req_similarity_pct ) {
					cdebug( 3, " > Failed edesc check, similarity is " + simdex );				
					is_bad = true;
				}
			}
			
			if( ! current.terrain.equals( "" ) && ! current.terrain.equals( r.terrain ) ) {
				cdebug( 3, " > Failed terrain check" );
				is_bad = true;
			}
			
			// name and "similar"
			if( ! current.name.equals( "" ) ) {
				// Do similar
				float simdex = metrics.getSimilarity( current.name, r.name );
				cdebug( 3, "Name similarity is: " + simdex );
				
				if( simdex < req_similarity_pct ) {
					cdebug( 3, " > Failed name check, similarity is " + simdex );				
					is_bad = true;
				}
			}
			
			if( got_exits ) {
				// exits, this idea is taken from azazellos mapper and it's a pretty good one
				// basically we start with a score of 0, and go exit by exit.  If
				// there is an exit here and an exit in the test room, we add 1.
				// if there is no exit here and and exit in the test room, we only
				// add 1 if there is a door in the test room too, so that hiddens
				// don't screw us.  Things like trolls digging and stuff do screw the
				// algo, but it's good enough and sane.  In the end, if the score
				// is not 6 (each exit passed ok) then we discard the room
				int score = 0;
				if( validate_exit( current.exit_e, current.door_e, r.exit_e, r.door_e ) ) score++;
				if( validate_exit( current.exit_w, current.door_w, r.exit_w, r.door_w ) ) score++;
				if( validate_exit( current.exit_n, current.door_n, r.exit_n, r.door_n ) ) score++;
				if( validate_exit( current.exit_s, current.door_s, r.exit_s, r.door_s ) ) score++;
				if( validate_exit( current.exit_u, current.door_u, r.exit_u, r.door_u ) ) score++;
				if( validate_exit( current.exit_d, current.door_d, r.exit_d, r.door_d ) ) score++;
				
				if( score != 6 ) {
					cdebug( 3, " > Failed exit check, score " + score );
					is_bad = true;
				}
			}

			if( is_bad ) {
				potential_rooms.remove( i );
				i--; // do -- because all items move -1
			}
		}
		
		// If only 1 room remains we must be there
		if( potential_rooms.size() == 1 && last_move_unexplored == false ) {
			last_potential_rooms = null;
			last_room = curr_room;
			curr_room = (CheeseRoomData)potential_rooms.get( 0 );
			x = curr_room.x;
			y = curr_room.y;
			z = curr_room.z;
			//set_room( (CheeseRoomData)potential_rooms.get( 0 ) );
			cdebug( 2, "Located self at room " + curr_room.roomid );
			
			// Print hiddens
			print_hidden_exit( curr_room.door_n, "n" );
			print_hidden_exit( curr_room.door_s, "s" );
			print_hidden_exit( curr_room.door_e, "e" );
			print_hidden_exit( curr_room.door_w, "w" );
			print_hidden_exit( curr_room.door_u, "u" );
			print_hidden_exit( curr_room.door_d, "d" );
		}else{
			// Just mod our x/y/z to where we would go based on the last dir we went,
			// so that mapping can happen in new areas
			cdebug( 2, "Could not find me, have " + potential_rooms.size() + " potentials for next" );
			//set_room( null );
			last_room = curr_room;
			
			if( potential_rooms.size() == 1 ) {
				// this means we moved through an unexplored exit, set curr room but later
				// override x/y/z to new "logical" location.  If they merge later
				// then it'll all come out ok anyway
				curr_room = (CheeseRoomData)potential_rooms.remove( 0 );
			}else{
				curr_room = null;
			}
			
			if( potential_rooms.size() > 0 )
				// save these rooms for next time
				last_potential_rooms = potential_rooms;
			else
				last_potential_rooms = null;
			
			// X/Y work like when you draw to the screen,
			// X runs left to right getting bigger,
			// Y runs top to bottom getting bigger
			switch( last_move ) {
				case DIR_NORTH:
					y--;
					break;
				
				case DIR_SOUTH:
					y++;
					break;
					
				case DIR_EAST:
					x++;
					break;
				
				case DIR_WEST:
					x--;
					break;
				
				case DIR_UP:
					z++;
					break;
				
				case DIR_DOWN:
					z--;
					break;
					
				default:
					// hmm dunno
					break;
			}
		}
		
		// save player to do update
		save_player();
		
		if( automerge && curr_room != null && last_room != null
				&& curr_room.roomid != last_room.roomid ) // do not auto-link to self
			 cmd_save( "@save merge" );
		
		// Be VERY careful using this
		if( autoforce && curr_room != null )
			cmd_save( "@save force" );
		
		// Handle autosave
		if( autosave && curr_room == null )
			cmd_save( "@save" );
			
		
		// If we are found, then save auto-room-color
		if( room_color == null ) {
			if( curr_room != null ) {
				room_color = last_auto_room_color;
				cdebug( 3, "Locking on room color: " + room_color );
			}
		}
	}
	
	private boolean validate_exit( int curr_exit, String curr_door,
			int test_exit, String test_door ) {
		// key as 1 == exit, 0 == no exit, -1 == door
		int curr = 0;
		if( curr_exit != 0 )
			curr = 1;
		if( ! curr_door.equals( "" ) )
			curr = -1;
		
		int test = 0;
		if( test_exit != 0 )
			test = 1;
		if( ! test_door.equals( "" ) ) 
			test = -1;
		
		// now if both are 1 or both are 0 we are ok
		// otherwise if test == -1 then it may be a hidden door
		// or broken door if we don't see it in curr
		if( test == curr )
			return( true );
		
		if( test == -1 )
			return( true );
		
		CheeseLab.dmesg( 4, "Exit failed: " + curr_exit + "/" + curr_door + " " + test_exit + "/" + test_door );
		return( false );
	}
	
	/*
	private void set_room( CheeseRoomData r ) {
		if( r == null ) {
			CheeseLab.dmesg( "Setting null room" );
		}else{
			CheeseLab.dmesg( "Setting room to " + r.roomid );
		}
		
		last_room = curr_room;
		curr_room = r;

		CheeseLab.dmesg( "Last room is: " + ( last_room == null ? "NULL" : Integer.toString( last_room.roomid ) ) );
		CheeseLab.dmesg( "Curr room is: " + ( curr_room == null ? "NULL" : Integer.toString( curr_room.roomid ) ) );
		
		if( r != null ) {
			x = r.x;
			y = r.y;
			z = r.z;
		
			save_player();
		}
	}
	*/
	
	private void save_player() {
		// save player state to db and inc global state
		try {
			CheeseLab.dmesg( 2, "Saving player: " + username );
			PreparedStatement s = db.prepareStatement(
					"update user_t " + 
					" set x=?, y=?, z=?, roomid=?, " +
					" hit_curr=?, hit_max=?, " +
					" move_curr=?, move_max=?, " +
					" mana_curr=?, mana_max=?, " +
					" online=?, future=? " +
					"where userid=?" );
			s.setInt( 1, x );
			s.setInt( 2, y );
			s.setInt( 3, z );
			
			if( curr_room != null )
				s.setInt( 4, curr_room.roomid );
			else
				s.setInt( 4, 0 );
				
			s.setInt( 5, hits ); s.setInt( 6, hits_max );
			s.setInt( 7, moves ); s.setInt( 8, moves_max );
			s.setInt( 9, mana ); s.setInt( 10, mana_max );
			
			s.setInt( 11, online ? 1 : 0 );
			
			// Make our futures string based on the vectors
			String strfut = "";
			for( int i = 0; i < future.size(); i++ ) {
				if( i > 0 )
					strfut += "-";
				strfut += future.get( i );
			}
			//CheeseLab.dmesg( 0, "strfut: " + strfut );
			s.setString( 12, strfut );
			
			s.setString( 13, username );
			
			s.executeUpdate();
			s.close();
			CheeseLab.dmesg( 2, "Save complete" );
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		CheeseProxy.changed();
	}
	
	// save, save merge, save init, save force
	// merge => link 2 rooms together that already exist
	// init => create new room connected to nothing
	// force => save room here even though it exists in the db elsewhere
	private void cmd_save( String line ) {
		// Reload this and last rooms in case they have changed
		if( curr_room != null )
			curr_room = CheeseDecoder.get_room( curr_room.roomid, db );
		
		if( last_room != null )
			last_room = CheeseDecoder.get_room( last_room.roomid, db );
		
		
		// Verify we have enough data to save
		String[] split = line.split( " " );
		String how = "";
		if( split.length > 1 ) {
			how = split[ 1 ];
			CheeseLab.dmesg( 2, "Save '" + how + "'" );
		}
		
		if( current.edesc.equals( "" ) ) {
			CheeseLab.dmesg( 3, "No edesc" );
			return;
		}
		
		if( current.name.equals( "" ) ) {
			CheeseLab.dmesg( 3, "No name" );
			return;
		}

		if( ! got_exits ) {
			CheeseLab.dmesg( 3, "Never got exits line" );
			return;
		}
		
		// Modify what the save command sees based on args
		if( how.equals( "force" ) ) {
			// Clear this room
			curr_room = null;
		}
		
		if( how.equals( "merge" ) ) {
			// Merge requires we know where we are
			if( curr_room == null ) {
				CheeseLab.dmesg( 3, "Nothing to merge" );
				return;
			}
		}
		
		// handle init, where null last room is ok (required really)
		if( how.equals( "init" ) ) {
			last_room = null;
		}else if( last_room == null ) {
			println_player( "Unable to save: no last room to link to" );
			CheeseLab.dmesg( 3, "Error saving: no last room" );
			return;
		}
		
		// More than one room here, must force (dup rooms)
		if( last_potential_rooms != null && ! how.equals( "force" ) ) {
			println_player( "Multiple potential rooms here, must force to save" );
			CheeseLab.dmesg( 3, "Multiple duplicate rooms, must force to save" );
			return;
		}
		
		// catch we know where we are, must be merged or forced if dup room
		if( curr_room != null && ! how.equals( "merge" ) && ! how.equals( "force" ) ) {
			println_player( "Player is in a known location, merge this room or force if it's a duplicate" );
			CheeseLab.dmesg( 3, "I think I know where we are, merge or force" );
			return;
		}
		
		try {
			PreparedStatement s;	
			if( curr_room == null ) {
				println_player( "Saving a new room" );
				CheeseLab.dmesg( 2, "Saving a new room" );
				// save new room
				s = db.prepareStatement( "select roomid from room_t where x=? and y=? and z=?" );
				s.setInt( 1, x );
				s.setInt( 2, y );
				
				// Check x/y/z is ok to save here			
				boolean here_is_ok = false;
				while( ! here_is_ok ) {
					s.setInt( 3, z );
					ResultSet rs = s.executeQuery();
					if( rs.next() ) {
						// there is something here, suck
						z--;
					}else{
						here_is_ok = true;
					}
					rs.close();
				}
				s.close();
			
				// OK we should have all the stuff we need
				s = db.prepareStatement( 
						"insert into room_t ( name, edesc, terrain," +
						"	userid, mtime, overlays," +
						"	x, y, z," +
						"	exit_n, door_n," +
						"	exit_s, door_s," +
						"	exit_e, door_e," +
						"	exit_w, door_w," +
						"	exit_u, door_u," +
						"	exit_d, door_d )" +
						"values ( ?, ?, ?," +
						"   ?, ?, ''," +
						" 	?, ?, ?," +
						"	?, ?," +
						"	?, ?," +
						"	?, ?," +
						"	?, ?," +
						"	?, ?," +
						"	?, ?)" );
				int col = 1;
				s.setString( col++, current.name ); 
				s.setString( col++, current.edesc );
				s.setString( col++, current.terrain );
				
				// Now update this rooms mtime/username it will happen on *any* kind of change
				long now = System.currentTimeMillis();
				s.setString( col++, username ); s.setLong( col++, now );
				
				s.setInt( col++, x ); s.setInt( col++, y ); s.setInt( col++, z );
				
				s.setInt( col++, current.exit_n ); s.setString( col++, current.door_n );
				s.setInt( col++, current.exit_s ); s.setString( col++, current.door_s );
				s.setInt( col++, current.exit_e ); s.setString( col++, current.door_e );
				s.setInt( col++, current.exit_w ); s.setString( col++, current.door_w );
				s.setInt( col++, current.exit_u ); s.setString( col++, current.door_u );
				s.setInt( col++, current.exit_d ); s.setString( col++, current.door_d );
				s.executeUpdate();
				s.close();
				
				// Get the last and put us here
				s = db.prepareStatement( "select * from room_t where x=? AND y=? and z=? AND userid=? AND mtime=?" );
				s.setInt( 1, x );
				s.setInt( 2, y );
				s.setInt( 3, z );
				s.setString( 4, username );
				s.setLong( 5, now );
				ResultSet rs = s.executeQuery();
				if( rs.next() ) {
					CheeseRoomData room = CheeseDecoder.sql_to_room( rs );
					curr_room = room;
					save_player();
				}
				rs.close();
				s.close();
			}
			
			// Curr room is set now, even if we added it
			if( last_room != null ) {
				// Also attach the last room to here and here to the last room
				// which is technically not always correct (going back the way
				// you came may NOT take you where you came from) but in many
				// cases it is, and you can just fix the bad cases
				//
				// Also we only attach if the exit was unexplored
				String here_exit = null, 
					last_exit = null;
				int here_toid = 0,
					last_toid = 0;
				
				switch( last_move ) {
					case DIR_NORTH:
						here_exit = "exit_s";
						here_toid = curr_room.exit_s;
						last_exit = "exit_n";
						last_toid = last_room.exit_n;
						break;
						
					case DIR_SOUTH:
						here_exit = "exit_n";
						here_toid = curr_room.exit_n;
						last_exit = "exit_s";
						last_toid = last_room.exit_s;
						break;
						
					case DIR_EAST:
						here_exit = "exit_w";
						here_toid = curr_room.exit_w;
						last_exit = "exit_e";
						last_toid = last_room.exit_e;
						break;
						
					case DIR_WEST:
						here_exit = "exit_e";
						here_toid = curr_room.exit_e;
						last_exit = "exit_w";
						last_toid = last_room.exit_w;
						break;
						
					case DIR_UP:
						here_exit = "exit_d";
						here_toid = curr_room.exit_d;
						last_exit = "exit_u";
						last_toid = last_room.exit_u;
						break;
						
					case DIR_DOWN:
						here_exit = "exit_u";
						here_toid = curr_room.exit_u;
						last_exit = "exit_d";
						last_toid = last_room.exit_d;
						break;
				}
			
				// Setup a now that says we modded this
				long now = System.currentTimeMillis();
				PreparedStatement mark_changed = db.prepareStatement( "update room_t set mtime=?, userid=? where roomid=?" );
				mark_changed.setLong( 1, now );
				mark_changed.setString( 2, username );
				
				// connect exits
				if( here_toid == -1 ) {
					println_player( "Linking room this room " + here_exit + " to #" + last_room.roomid );
					
					s = db.prepareStatement( "update room_t set " + here_exit + "=? where roomid=?" );
					s.setInt( 1, last_room.roomid );
					s.setInt( 2, curr_room.roomid );
					s.executeUpdate();
					
					mark_changed.setInt( 3, curr_room.roomid );
					mark_changed.executeUpdate();
				}
			
				if( last_toid == -1 ) {
					println_player( "Linking room #" + last_room.roomid + " " + last_exit + " here" );
					s = db.prepareStatement( "update room_t set " + last_exit + "=? where roomid=?" );
					s.setInt( 1, curr_room.roomid );
					s.setInt( 2, last_room.roomid );
					s.executeUpdate();
					
					mark_changed.setInt( 3, last_room.roomid );
					mark_changed.executeUpdate();
				}
			}
			
			CheeseProxy.world_changed();
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}

	private void cmd_help() {
		println_player(	"CheeseLab MUME Mapping Tools v." + CheeseLab.version );
		println_player( "" );
			
		try {
			FileInputStream fstream = new FileInputStream( "proxyhelp.txt" );
			BufferedReader greet = new BufferedReader( new InputStreamReader( fstream ) );
			String line;
			while( (line = greet.readLine() ) != null ) {
				println_player( line );
			}
			fstream.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}
	

	private void cmd_stack() {
		println_player( "Current stack: " );
		println_player( " Name: " + current.name );
		println_player( " Edesc: " );
		println_player( current.edesc );
		println_player( "-" );
		println_player( " Current RoomID: " + (curr_room == null ? "UNKOWN" : Integer.toString( curr_room.roomid )) );
		println_player( " Last RoomID: " + (last_room == null ? "UNKOWN" : Integer.toString( last_room.roomid )) );
		println_player( "-" );
		
		String movestack_str = "";
		for( int i = 0; i < movestack.size(); i++ ) {
			if( i > 0 ) 
				movestack_str += ",";
			movestack_str += movestack.get( i );
		}
		println_player( " Movestack: " + movestack_str );
	}
	
	private void cmd_password( String line ) {
		String[] split = line.split( " " );
		String newpw = "";
		if( split.length > 1 )
			newpw = split[ 1 ];
		
		try {
			PreparedStatement s = db.prepareStatement( "update user_t set password=? where userid=?" );
			s.setString( 1, newpw );
			s.setString( 2, username );
			s.executeUpdate();
			
			println_player( "Changed password for " + username );
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}
	
	private void cmd_user( String line ) {
		String[] split = line.split( " " );
		if( split.length == 1 ) {
			println_player( "Invalid usage" );
			return;
		}
		
		String cmd = split[ 1 ];
		
		if( cmd.equals( "list" ) ) {
			// print users
			try {
				PreparedStatement s = db.prepareStatement( "select userid from user_t order by userid" );
				ResultSet rs = s.executeQuery();
				while( rs.next() ) {
					CheesePlayerData p = CheeseDecoder.get_player( rs.getString( "userid" ), db );
					println_player( "(" + p.permissions + ") " + p.name );
				}
				rs.close();
				s.close();
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}else if( cmd.equals( "create" ) ) {
			// this lets you create a new user
			try {
				String newuser = "";
				if( split.length > 2 )
					newuser = split[ 2 ];
				
				if( newuser.equals( "" ) ) {
					println_player( "Need username to create" );
				}else{
					PreparedStatement s = db.prepareStatement( "insert into user_t (userid, password) values (?,?)" );
					s.setString( 1, split[ 2 ] );
					s.setString( 2, s.toString() ); // supposed to be non-guessable :)
					s.executeUpdate();
					s.close();
					println_player( "Created: " + split[ 2 ] );					
				}
			}catch( Exception e ) {
				e.printStackTrace();
			}			
		}else if( cmd.equals( "password" ) ) {
			// change a users password
			String uid = "";
			String newpw = "";
			if( split.length > 3 ) {
				uid = split[ 2 ];
				newpw = split [ 3 ];
			}
			
			try {
				PreparedStatement s = db.prepareStatement( "update user_t set password=? where userid=?" );
				s.setString( 1, newpw );
				s.setString( 2, uid );
				s.executeUpdate();
				
				println_player( "Changed password for " + uid );
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}else if( cmd.equals( "delete" ) ) {
			try {
				String deluser = "";
				if( split.length > 2 )
					deluser = split[ 2 ];
				
				if( deluser.equals( "" ) ) {
					println_player( "Need user to delete" );
				}else{
					PreparedStatement s = db.prepareStatement( "delete from user_t where userid=?" );
					s.setString( 1, split[ 2 ] );
					s.executeUpdate();
					s.close();
					println_player( "Deleted: " + split[ 2 ] );
				}
			}catch( Exception e ) {
				e.printStackTrace();
			}			
		}else if( cmd.equals( "perm" ) ) {
			// save permissions
			String uid = "";
			String newperm = "";
			if( split.length > 3 ) {
				uid = split[ 2 ];
				newperm = split [ 3 ];
			}
			
			try {
				PreparedStatement s = db.prepareStatement( "update user_t set permissions=? where userid=?" );
				s.setString( 1, newperm );
				s.setString( 2, uid );
				s.executeUpdate();
				
				println_player( "Changed permissions for " + uid );
			}catch( Exception e ) {
				e.printStackTrace();
			}
		}else{
			println_player( "Unknown option" );
		}
	}
	
	private String translate_terrain( String terrain ) {
		// Translate terrain to the original terrain at time of
		// first writing this mapper, e.g. if the change tile types
		// we gotta map them right
		
		// This used to be in a file, but I don't think that's necessary since
		// this shouldn't change often
		
		// Indoors
		if( terrain.equals( "[" ) )
			return( "@" );

		// Shallow water
		if( terrain.equals( "%" ) )
			return( "~" );
		
		if( terrain.equals( ":" ) )
			return( "\"" );

		if( terrain.equals( "." ) )
			return( "_" );
		
		if( terrain.equals( "f" ) )
			return( "%" );

		if( terrain.equals( "~" ) )
			return( "w" );		
		
		// Not something they messed up
		return( terrain );
	}
	
	public void println_player( String s ) {
		print_player( s + "\r\n" );
	}

	
	public void print_player( String s ) {
		print_player_raw( "## " + s );
	}
	
	public void print_player_raw( String s ) {			
		ByteBuffer outbuffer = ByteBuffer.allocate( s.length() );
		try {
			outbuffer.clear();
			outbuffer.put( s.getBytes( "ISO-8859-1" ) );
			outbuffer.flip();
			sock.write( outbuffer );
		}catch( Exception e ) {
			e.printStackTrace();
		}
	}

	// These methods are for using at the login process mainly, they shouldn't
	// be used ANYWHERE else
	public byte[] readline_player_bytes() {
		ByteBuffer b = ByteBuffer.allocate( 1024 );
		b.clear();
		try {
			sock.read( b );
		}catch( Exception e ) {
			e.printStackTrace();
		}
		
		b.flip();
		return( next_raw_line( b ) );
	}
	
	public String readline_player() {
		return( new String( strip_line( readline_player_bytes() ) ) );
	}
	
	private void print_hidden_exit( String exit, String dir ) {
		if( exit.equals( "" ) )
			return;
		
		if( exit.equals( "exit" ) )
			return;
		
		println_player( "Door " + dir + ": " + exit );
	}
	
	public void cdebug( int level, String s ) {
		if( level <= debuglevel ) {
			println_player( " DEBUG " + level + " "+ s );
		}
		CheeseLab.dmesg( level, s );
	}
	
	public String enum_to_dir( int dir ) {
		switch( dir ) {
			case DIR_NORTH:
				return( "north" );
			
			case DIR_SOUTH:
				return( "south" );
			
			case DIR_EAST:
				return( "east" );
			
			case DIR_WEST:
				return( "west" );
			
			case DIR_UP:
				return( "up" );
			
			case DIR_DOWN:
				return( "down" );
		}
		
		return( "unknown" );
	}

}

